#  Copyright (C) 2016-2019  Kyle.Hwang <upday7[at]163.com>
#  #
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version, with the additions
#  listed at the end of the accompanied license file.
#  #
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#  #
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#  #
#  NOTE: This program is subject to certain additional terms pursuant to
#  Section 7 of the GNU Affero General Public License.  You should have
#  received a copy of these additional terms immediately following the
#  terms and conditions of the GNU Affero General Public License which
#  accompanied this program.



import json
from urllib.parse import quote as urlquote
from urllib.request import urlopen

UD_DEFID_URL = 'https://api.urbandictionary.com/v0/define?defid='
UD_DEFINE_URL = 'https://api.urbandictionary.com/v0/define?term='
UD_RANDOM_URL = 'https://api.urbandictionary.com/v0/random'


class UrbanDefinition(object):
    def __init__(self, word, definition, example, upvotes, downvotes):
        self.word = word
        self.definition = definition
        self.example = example
        self.upvotes = upvotes
        self.downvotes = downvotes

    def __str__(self):
        return '%s: %s%s (%d, %d)' % (
            self.word,
            self.definition[:50],
            '...' if len(self.definition) > 50 else '',
            self.upvotes,
            self.downvotes
        )


def _get_urban_json(url):
    f = urlopen(url)
    data = json.loads(f.read().decode('utf-8'))
    f.close()
    return data


def _parse_urban_json(json, check_result=True):
    result = []
    if json is None or any(e in json for e in ('error', 'errors')):
        raise Exception('UD: Invalid input for Urban Dictionary API')
    if check_result and ('list' not in json or len(json['list']) == 0):
        return result
    for definition in json['list']:
        d = UrbanDefinition(
            definition['word'],
            definition['definition'],
            definition['example'],
            int(definition['thumbs_up']),
            int(definition['thumbs_down'])
        )
        result.append(d)
    return result


def define(term):
    """Search for term/phrase and return list of UrbanDefinition objects.

    Keyword arguments:
    term -- term or phrase to search for (str)
    """
    json = _get_urban_json(UD_DEFINE_URL + urlquote(term))
    return _parse_urban_json(json)


def defineID(defid):
    """Search for UD's definition ID and return list of UrbanDefinition objects.

    Keyword arguments:
    defid -- definition ID to search for (int or str)
    """
    json = _get_urban_json(UD_DEFID_URL + urlquote(str(defid)))
    return _parse_urban_json(json)


def random():
    """Return random definitions as a list of UrbanDefinition objects."""
    json = _get_urban_json(UD_RANDOM_URL)
    return _parse_urban_json(json, check_result=False)



#  Copyright (C) 2016-2019  Kyle.Hwang <upday7[at]163.com>
#  #
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version, with the additions
#  listed at the end of the accompanied license file.
#  #
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#  #
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#  #
#  NOTE: This program is subject to certain additional terms pursuant to
#  Section 7 of the GNU Affero General Public License.  You should have
#  received a copy of these additional terms immediately following the
#  terms and conditions of the GNU Affero General Public License which
#  accompanied this program.
import re
from abc import abstractmethod
from typing import Type

from .exception import NoTranslationSegmentError
from .prviders.base_provider import BaseProvider
from .prviders.cn_bing import CNBing
from .prviders.cn_bing_s import CNBingSuggestion
from .prviders.collinsdictionary import CollinsWeb
from .prviders.lexico import Lexico
from .prviders.spanishdict import SpanishDict
from .prviders.urbandictionary import define
from .prviders.vocabulary_s import VocabularySuggestion
from .prviders.vocaublary import Vocabulary


class DictionaryFactory:

    def __init__(self, dict_cls: Type[BaseProvider], word: str = ''):
        self.dict_cls = dict_cls
        self.word = word
        self._in_lang = None
        self._target_lang = None

    def __enter__(self, ):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        ...

    @property
    @abstractmethod
    def segment(self) -> str:
        ...

    @property
    @abstractmethod
    def lang_codes(self) -> dict:
        """
        {
            'es': 'spanish',
            'zh': 'chinese',
            'de': 'german',
            'fr': 'french',
            'en': 'english',
        }
        :return:
        """

    @property
    def provider(self) -> BaseProvider:
        return self.dict_cls(self.word, self.segment)

    def search(self, word: str, in_lang: str, target_lang: str) -> dict:
        self.word = word if word else self.word
        self._in_lang = in_lang
        self._target_lang = target_lang
        return self.provider.to_dict()


class OrphanDictionaryFactory(DictionaryFactory):
    @property
    def segment(self) -> str:
        return ''

    @property
    def lang_codes(self) -> dict:
        return {}

    def __init__(self, provider_cls, *, word: str = ''):
        self._provider_cls = provider_cls
        super(OrphanDictionaryFactory, self).__init__(self._provider_cls, word=word)

    def do_search(self, word: str = ""):
        return super(OrphanDictionaryFactory, self).search(word, '', '')


class CollinsDictionary(DictionaryFactory):
    lang_codes = {
        'es': 'spanish',
        'zh': 'chinese',
        'de': 'german',
        'fr': 'french',
        'en': 'english',
    }

    @property
    def segment(self) -> str:
        if self._in_lang not in self.lang_codes or self._target_lang not in self.lang_codes:
            raise NoTranslationSegmentError(str(self._provider_cls),
                                            self._in_lang, self._target_lang)
        if self._in_lang != 'en' and self._target_lang != 'en':
            raise NoTranslationSegmentError(str(self._provider_cls),
                                            self._in_lang, self._target_lang)
        if self._in_lang == self._target_lang == 'en':
            return 'english'
        return f"{self.lang_codes[self._in_lang]}-{self.lang_codes[self._target_lang]}"

    def __init__(self, *, word: str = ''):
        self._provider_cls = CollinsWeb
        super(CollinsDictionary, self).__init__(self._provider_cls, word=word)

    def en2fr(self, word: str = ''):
        return self.search(word, 'en', 'fr')

    def en2es(self, word: str = ''):
        return self.search(word, 'en', 'es')

    def en2de(self, word: str = ''):
        return self.search(word, 'en', 'de')

    def en2zh(self, word: str = ''):
        return self.search(word, 'en', 'zh')

    def es2en(self, word: str = ''):
        return self.search(word, 'es', 'en')

    def fr2en(self, word: str = ''):
        return self.search(word, 'fr', 'en')

    def de2en(self, word: str = ''):
        return self.search(word, 'de', 'en')

    def zh2en(self, word: str = ''):
        return self.search(word, 'zh', 'en')

    def en(self, word: str = ''):
        return self.search(word, 'en', 'en')


class OxfordDictionary(DictionaryFactory):
    lang_codes = {
        'es': 'spanish',
        'en': 'english',
    }

    @property
    def segment(self) -> str:
        if self._in_lang not in self.lang_codes or self._target_lang not in self.lang_codes:
            raise NoTranslationSegmentError(str(self._provider_cls),
                                            self._in_lang, self._target_lang)
        if self._in_lang == self._target_lang:
            return self._in_lang

        return f"{self._in_lang}-{self._target_lang}"

    def __init__(self, *, word: str = ''):
        self._provider_cls = Lexico
        super(OxfordDictionary, self).__init__(self._provider_cls, word=word)

    def en2es(self, word: str = ''):
        return self.search(word, 'en', 'es')

    def es2en(self, word: str = ''):
        return self.search(word, 'es', 'en')

    def en(self, word: str = ''):
        return self.search(word, 'en', 'en')

    def es(self, word: str = ''):
        return self.search(word, 'es', 'es')


class VocabularyDictionary(OrphanDictionaryFactory):

    def __init__(self, *, word: str = ''):
        super(VocabularyDictionary, self).__init__(Vocabulary, word=word)


class VocabularySuggestionDictionary(OrphanDictionaryFactory):

    def __init__(self, *, word: str = ''):
        super(VocabularySuggestionDictionary, self).__init__(VocabularySuggestion, word=word)


class SpanishDictDictionary(OrphanDictionaryFactory):

    def __init__(self, *, word: str = ''):
        super(SpanishDictDictionary, self).__init__(SpanishDict, word=word)


class CNBingDictionary(OrphanDictionaryFactory):
    def __init__(self, *, word: str = ''):
        super(CNBingDictionary, self).__init__(CNBing, word=word)


class CNBingSuggestionDictionary(OrphanDictionaryFactory):
    def __init__(self, *, word: str = ''):
        super(CNBingSuggestionDictionary, self).__init__(CNBingSuggestion, word=word)


class UrbanDictionary:

    def __init__(self, word: str = ''):
        self.word = word

    def do_search(self, word: str = ''):
        if word:
            self.word = word
        ub = define(self.word)
        return {
            'definitions': [
                re.sub(r"\s+", " ", d.definition.strip().replace("]", "").replace("[", "")) for d in ub]
        }



#  Copyright (C) 2016-2019  Kyle.Hwang <upday7[at]163.com>
#  #
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version, with the additions
#  listed at the end of the accompanied license file.
#  #
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#  #
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#  #
#  NOTE: This program is subject to certain additional terms pursuant to
#  Section 7 of the GNU Affero General Public License.  You should have
#  received a copy of these additional terms immediately following the
#  terms and conditions of the GNU Affero General Public License which
#  accompanied this program.

import re

from .base_provider import BaseProvider
from ..parser import Parser


class _BriefDefProvider(Parser):

    def val_pos(self):
        return self.select("span.pos")

    def val_def(self):
        return self.select("span.def > span")


class _AuthDefSegProvider(Parser):

    def val_topic(self):
        return [{
            'en': ent.text,
            'zh': zht.text
        } for ent, zht in zip(
            self.select("span.val_dis", one=False, text=False),
            self.select("span.bil_dis", one=False, text=False),
        )]

    def val_defs(self):
        return [{
            'en': ent.text,
            'zh': zht.text
        } for ent, zht in zip(
            self.select("div.def_pa", text=False).find_all('span', class_='val'),
            self.select("div.def_pa", text=False).find_all('span', class_='bil'),
        )]

    def val_examples(self):
        return [{
            'en': ent.text,
            'zh': zht.text
        } for ent, zht in zip(
            self.select("div.li_exs", text=False).find_all('div', class_='val_ex'),
            self.select("div.li_exs", text=False).find_all('div', class_='bil_ex'),
        )]


class _AuthDefProvider(Parser):
    def val_pos(self):
        return self.select("div.pos", )

    def val_def(self):
        return self.provider_to_list(_AuthDefSegProvider, "div.de_seg")

    def val_idm(self):
        idm_tag = self.bs.find('div', class_='idm_seg')
        if not idm_tag:
            return []
        try:
            return [
                {
                    'idm': idm.text,
                    'example': {
                        'en': ent.text,
                        'zh': zht.text
                    }
                } for idm, ent, zht in
                zip(
                    idm_tag.find_all('span', class_='ids'),
                    idm_tag.find_all('span', class_='val'),
                    idm_tag.find_all('span', class_='bil'),
                )
            ]
        except AttributeError:
            return []


class _EnZhDefProvider(Parser):

    def val_pos(self):
        return self.select("div.pos.pos1")

    def val_defs(self):
        return list(set(self.select("div.df_cr_w", one=False)))


class _EnEnDefProvider(Parser):

    def val_pos(self):
        return self.select("div.pos.pos1")

    def val_defs(self):
        return list(set(self.select("div.df_cr_w", one=False)))


class _DefsProvider(Parser):

    def val_auth(self):
        return self.provider_to_list(_AuthDefProvider, "div.each_seg")

    def val_ec(self):
        return self.provider_to_list(_EnZhDefProvider, 'div#crossid > table > tr')

    def val_ee(self):
        return self.provider_to_list(_EnEnDefProvider, 'div#homoid > table > tr')


class CNBing(BaseProvider):

    @property
    def url(self):
        return f"https://cn.bing.com/dict/search?q={self.word}"

    def __init__(self, word: str, seg=''):
        super(CNBing, self).__init__(word, seg)
        self.seg = seg

    @property
    def head_word(self):
        return self.select("div.hd_div > h1 > strong")

    @property
    def val_phone(self):
        p = re.compile(r"\[(?P<c>.+)\]")
        return {
            'uk': re.search(p, self.select("div.hd_pr")).group("c"),
            "us": re.search(p, self.select("div.hd_prUS")).group("c")
        }

    @property
    def val_images(self):
        return [t['src'].replace("&w=80&h=80", "")
                for t in self.select("div.img_area > div.simg > a > img", one=False, text=False)]

    @property
    def val_brief(self):
        return self.provider_to_list(_BriefDefProvider, ('li', {}), find_in_tag=self.bs.find('div', class_='qdef'))

    @property
    def val_defs(self):
        return self.provider_to_list(_DefsProvider, "div#defid")



#  Copyright (C) 2016-2019  Kyle.Hwang <upday7[at]163.com>
#  #
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version, with the additions
#  listed at the end of the accompanied license file.
#  #
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#  #
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#  #
#  NOTE: This program is subject to certain additional terms pursuant to
#  Section 7 of the GNU Affero General Public License.  You should have
#  received a copy of these additional terms immediately following the
#  terms and conditions of the GNU Affero General Public License which
#  accompanied this program.

import json
import re

from .base_provider import BaseProvider


class SpanishDict(BaseProvider):

    @property
    def url(self):
        return f'https://www.spanishdict.com/translate/{self.word}'

    def __init__(self, word: str, seg='es-en'):
        super(SpanishDict, self).__init__(word, seg)
        self.seg = seg

    @property
    def json(self) -> dict:
        return json.loads(re.search(
            'SD_DICTIONARY_RESULTS_PROPS\s?=(?P<c>.+);.+global.SD_WORD_ROOT_PROPS.+',
            self.rsp.content.decode().replace("\n", ""), re.MULTILINE
        ).group(1))['es']['entry']

    @property
    def head_word(self):
        return self.select("div#headword-es")

    @property
    def val_audio(self):
        return f'http://audio1.spanishdict.com/audio?lang=es&text={self.head_word}'

    @property
    def val_neodict(self):
        return self.json['neodict']
